#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .mbio import MBIO

from .items import Items
from .config import MBIOConfig

import time

# from prettytable import PrettyTable
from digimat.units import Units


FLAG_ERROR = (0x1 << 0)
FLAG_MANUAL = (0x1 << 1)
FLAG_OVERRIDE = (0x1 << 2)
FLAG_DISABLED = (0x1 << 3)
FLAG_REMOTEMANUAL = (0x1 << 4)
FLAG_COMMISSIONABLE = (0x1 << 5)
FLAG_MANAGED = (0x1 << 6)
FLAG_WRITABLE = (0x1 << 7)

# READ PRINCIPLE (values refresh)
# manager->refresh() retrieve I/O state and call value.updateValue(state)
# will set value._value (value.value)
# and call value.notify() if needed (queue values updates notifications in MBIO for dispatching)

# WRITE PRINCIPLE (value change orders)
# value.value=state or value.set(state)
# will set value._toReachValue (value.toreachValue) and call value.signalSync()
# manager->sync() checks value.isPendingSync(), update I/O and call value.clearSync()
# Note: a write don't update the value.value, which will be updated by the refresh process.
# If the value isn't auto refreshed, use value.clearSyncAndUpdateValue() instead of .clearSync()
# In the manager->sync() a device->signalRefresh() is automatically called after a manager->sync() to
# speedup the value refresh

# FIXME: the value managed attribute (init) means that it's .manager() has to be called
# This has nothing to do with the .isManaged() flag that indicates if this value is actively written by the CPU
# Kind of name conflict here


class MBIOValue(object):
    def __init__(self, parent: MBIOValues, name, unit=None, default=None, resolution=0.1, writable=False, commissionable=False):
        assert(isinstance(parent, MBIOValues))
        self._parent=parent
        if not name:
            name='%d' % parent.count()
        self._name=name
        self._key=parent.computeValueKeyFromName(name)
        self._writable=writable
        self._enable=True
        if type(unit)==str:
            unit=parent.units.getByName(unit)
        self._unit=unit
        self._error=False
        self._manual=None
        self._override=False
        self._remotemanual=False
        self._commissionable=commissionable
        self._stamp=0
        self._min=None
        self._max=None
        self._default=self.normalizeValue(default)
        self._value=self._default
        self._flags=None
        self._config=MBIOConfig()
        self._resolution=resolution
        self._enableNotify=True
        self._lastValueNotified=None
        self._timeoutNotifyRefresh=0
        self._timeoutManaged=0
        self._notifyCount=0
        self.onInit()
        self.updateFlags()
        self._parent.add(self)
        if self._default is not None:
            self.updateValue(self._default)

    def hasManager(self):
        return False

    def manager(self):
        # to be overriden for managed values
        self.logger.warning("MANAGER %s" % self)
        pass

    def timeout(self, delay):
        return time.time()+delay

    def checkIfValueWillTriggerNotify(self, v):
        if v is not None:
            try:
                if type(self._value) is bool:
                    if v != self._value:
                        return True
                    return False
                if self._resolution is not None:
                    if abs(self._value-v)>=self._resolution:
                        return True
                    return False
                if v!=self._value:
                    return True
            except:
                pass
        return False

    def checkNotify(self):
        if not self._enableNotify:
            return False
        try:
            if type(self._value) is bool or self._lastValueNotified is None:
                return True
            if self._resolution is not None:
                if abs(self._value-self._lastValueNotified)>=self._resolution:
                    return True
                return False
        except:
            pass
        return True

    @property
    def type(self):
        return self.__class__.__name__

    def notify(self, force=False, delayAutoRefresh=0):
        if self._enableNotify and force or self.checkNotify():
            self.parent.getMBIO().signalValueUpdateNotify(self)
            self._lastValueNotified=self.value
            if delayAutoRefresh>0:
                self._timeoutNotifyRefresh=self.timeout(delayAutoRefresh)
            else:
                self._timeoutNotifyRefresh=self.timeout(30)

    def notifyManager(self, delayAutoRefresh=0):
        if self._value is not None:
            if time.time()>=self._timeoutNotifyRefresh:
                self.notify(True, delayAutoRefresh)

    def enableNotify(self, state=True):
        self._enableNotify=state

    @property
    def notifyCount(self):
        return self._notifyCount

    @property
    def config(self) -> MBIOConfig:
        return self._config

    @property
    def parent(self) -> MBIOValues:
        return self._parent

    @property
    def logger(self):
        return self._parent.logger

    @property
    def key(self):
        return self._key

    @property
    def tag(self):
        return self.config.iomaptag

    def keyparts(self, part=None):
        try:
            parts=self.key.split('_')
            if part is None:
                return parts
            return parts[parts[part]]
        except:
            pass

    def keyrootpart(self):
        key=self.key
        size=len(self.parent.prefix)
        return key[:size+1]

    def keyleafpart(self):
        return self.name

    @property
    def name(self):
        return self._name

    def onInit(self):
        pass

    def setRange(self, vmin, vmax):
        self._min=vmin
        self._max=vmax

    @property
    def resolution(self):
        return self._resolution

    @resolution.setter
    def resolution(self, value):
        self._resolution = value

    def normalizeValue(self, value):
        if value is not None:
            # self.logger.warning(value)
            value=float(value)
            if self._min is not None:
                value=max(self._min, value)
            if self._max is not None:
                value=min(self._max, value)
        return value

    def isWritable(self):
        if self._writable:
            return True
        return False

    def isPendingSync(self, reset=False):
        return False

    def isError(self):
        return self._error

    def isManual(self):
        if self._manual is not None:
            return True
        return False

    def isManaged(self):
        """Return True if the value is writable and seems to be managed (value written not too long ago)"""
        return False

    def isRemoteManual(self):
        if self._remotemanual:
            return True
        return False

    def isCommissionable(self):
        if self._commissionable:
            return True
        return False

    def signalSync(self, value=None):
        pass

    def signalNotified(self):
        """The value has been sent to the notifiers (CPU)"""
        pass

    def manual(self, value=None):
        if value is None:
            value=self.value
        if value is not None:
            value=self.normalizeValue(value)
            if value!=self._manual:
                self._manual=value
                self.updateFlags()
            if self.isWritable():
                self.signalSync(self._manual)
            else:
                self.updateValue(self._manual)
                self._stamp=time.time()

    def auto(self):
        if self._manual is not None:
            self._manual=None
            self.updateFlags()

    @property
    def override(self):
        return self._override

    def setOverride(self, state=True):
        state=bool(state)
        if state!=self._override:
            self._override=state
            self.updateFlags()

    def setRemoteManual(self, state=True):
        state=bool(state)
        if state!=self._remotemanual:
            self._remotemanual=state
            self.updateFlags()

    def setError(self, state=True):
        state=bool(state)
        if state!=self._error:
            self._error=state
            self.updateFlags()

    def isEnabled(self):
        return self._enable

    def enable(self, state=True):
        state=bool(state)
        if state!=self._enable:
            self._enable=state
            self.updateFlags()

    def disable(self):
        self.enable(False)

    def age(self):
        return time.time()-self._stamp

    def isTimeout(self, delay):
        if self.age()>=delay:
            return True
        return False

    @property
    def value(self):
        if self.isError() and self._default is not None:
            return self._default
        return self._value

    def valuestr(self, sigdigits=1):
        if self.value is None:
            return None
        if self.isDigital():
            if self.value:
                return 'ON'
            return 'OFF'
        try:
            decimal=float(self.resolution) % 1
            if abs(decimal)>0:
                r=str(decimal)
                dcount=len(r[r.find('.')+1:])
                fmt='%%.0%df%%s' % dcount
            else:
                fmt='%.0f%s'
            return fmt % (self.value, self.unitstr())
        except:
            self.logger.exception('FMT %s' % fmt)
            pass
        try:
            return str(self.value)+self.unitstr()
        except:
            pass
        return str(self.value)

    def richvaluestr(self, sigdigits=1):
        if self.value is None:
            return "[bold blue]None[/bold blue]"
        if self.isDigital():
            if self.value:
                return '[bold green]ON[/bold green]'
            return '[bold red]OFF[/bold red]'
        try:
            decimal=float(self.resolution) % 1
            if abs(decimal)>0:
                r=str(decimal)
                dcount=len(r[r.find('.')+1:])
                fmt='[bold magenta]%%.0%df%%s[/bold magenta]' % dcount
            else:
                fmt='[bold magenta]%.0f%s[/bold magenta]'
            return fmt % (self.value, self.unitstr())
        except:
            self.logger.exception('FMT %s' % fmt)
            pass
        try:
            return str(self.value)+self.unitstr()
        except:
            pass
        return str(self.value)

    @property
    def flags(self):
        return self._flags

    def isFlag(self, flag):
        try:
            for f in flag:
                if f.upper() in self._flags:
                    return True
        except:
            pass
        return False

    def flagsAsValue(self):
        value=0
        if self.isFlag('E'):
            value |= FLAG_ERROR
        if self.isFlag('*'):
            value |= FLAG_OVERRIDE
        if self.isFlag('M'):
            value |= FLAG_MANUAL
        if self.isFlag('X'):
            value |= FLAG_DISABLED
        if self.isWritable():
            value |= FLAG_WRITABLE
            if self.isManaged():
                value |= FLAG_MANAGED
        if self.isFlag('C'):
            value |= FLAG_COMMISSIONABLE
        if self.isFlag('R'):
            value |= FLAG_REMOTEMANUAL

        return value

    def updateFlags(self):
        flags=''
        if self.override:
            flags+='*'
        if self.isError():
            flags+='E'
        if self.isManual():
            flags+='M'
        if self.isCommissionable():
            flags+='C'
        if not self.isEnabled():
            flags+='X'
        if self.isRemoteManual():
            flags+='R'

        if self.isWritable():
            flags+='W'
            if not self.isManaged():
                flags+='!'

        if flags!=self._flags:
            self._flags=flags
            # Force notify
            self.notify(True)

            # A flag change (as REMOTEMANUAL) may have an impact
            # on the way the value is processed (i.e. AnalogOutput delays)
            # So request a sync
            self.signalSync()
            return True

        return False

    def updateValue(self, value, unit=None):
        if self.isManual():
            if not self.isWritable():
                value=self._manual
        else:
            self._stamp=time.time()
        value=self.normalizeValue(value)
        notify=self._lastValueNotified is None
        if unit is not None and self._unit!=unit:
            self._unit=unit
            notify=True
        if value!=self._value:
            self._value=value
            notify=True
        if notify:
            self.notify()

    def updateValidValue(self, value, unit=None):
        self.updateValue(value, unit)
        self.setError(False)

    def isValid(self):
        if self.isEnabled() and self._value is None:
            return False
        return True

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        if type(value)==str:
            value=self.parent.units.getByName(value)
        if self.isDigital():
            value=15
        if self._unit!=value:
            self._unit=value
            self.notify()

    def unitstr(self):
        return self.parent.units.str(self.unit)

    def isDigital(self):
        # FIXED: a value is digital only if the object is a MBIODigitalValue instance
        # and not by it's unit. Otherwise, in case an analog value is falsy written with a digital unit,
        # the value object tends to be locked in a digital value.
        # if self.parent.units.isDigital(self.unit):
            # return True
        return False

    def isMultistate(self):
        return False

    def __bool__(self):
        self.logger.warning('WARNING: *************************** %s used as bool' % self)
        try:
            if self.isDigital():
                return self.isOn()
        except:
            pass
        if self.value!=0:
            return True
        return False

    def __repr__(self):
        if self.updateFlags():
            # should not be called
            self.logger.warning('%s --> updateFlags TRUE', self.key)
        age=int(self.age())
        if age>3600:
            age='N/A'
        if self._flags:
            return '%s(%s=%s, flags=[%s], %ss, #%d, tag=%s)' % (self.__class__.__name__, self.key,
                self.valuestr(), self._flags, age, self._notifyCount, self.tag)
        return '%s(%s=%s, %ss, #%d, tag=%s)' % (self.__class__.__name__, self.key, self.valuestr(), age, self._notifyCount, self.tag)

    def richstr(self):
        self.updateFlags()
        age=int(self.age())
        if age>3600:
            age='N/A'
        if self._flags:
            color='[white]'
            if self.isFlag('AE'):
                color='[red]'
            if self.isFlag('MR*'):
                color='[green]'
            if self.isFlag('X'):
                color='[bright_black]'

            return '%s%s[/]([bold]%s[/bold]=%s, flags=[[bold]%s[/bold]], %ss, #%d, tag=%s)' % (color, self.__class__.__name__, self.key,
                self.richvaluestr(), self._flags, age, self._notifyCount, self.tag)
        return '%s([bold]%s[/bold]=%s, %ss, #%d, tag=%s)' % (self.__class__.__name__, self.key, self.richvaluestr(), age, self._notifyCount, self.tag)

    def refpath(self):
        path=[]
        ref=self
        while True:
            try:
                ref=ref.parent
            except:
                break
            if isinstance(ref, MBIOValues):
                continue
            if ref is None:
                break
            path.insert(0, str(ref))
        return '->'.join(path)

    def match(self, key):
        try:
            if key is None:
                return True
            if key.lower() in self._key:
                return True
            if key=='*':
                return True
            if key[0]=='*' and key[-1]:
                key=key[1:-1].lower()
                if key in self._key:
                    return True
            if key[0]=='*':
                key=key[1:].lower()
                if self._key[-len(key):]==key:
                    return True
            if key[-1]=='*':
                key=key[:-1].lower()
                if self._key[0:len(key)]==key:
                    return True
        except:
            pass
        return False


class MBIOConstant(MBIOValue):
    def __init__(self, parent: MBIOValues, name, value, unit):
        super().__init__(parent, name=name, unit=unit, default=value)


# FIXME: duplicated code MBIOValueWritable - MBIOValueDigitalWritable
class MBIOValueWritable(MBIOValue):
    def __init__(self, parent: MBIOValues, name, unit=None, default=None, resolution=0.1, commissionable=False):
        super().__init__(parent, name=name, unit=unit, default=default, writable=True, resolution=resolution, commissionable=commissionable)
        self._toReachValue=None
        self._pendingSync=False

    def isManaged(self):
        """Return True if the value is writable and seems to be managed (value written not too long ago)"""
        if self.isValid() and time.time()<self._timeoutManaged:
            return True
        return False

    @MBIOValue.value.setter
    def value(self, value):
        if value is not None:
            if self.isEnabled():
                if self.isManual():
                    value=self._manual
                value=self.normalizeValue(value)
                resolution=self._resolution
                if resolution is not None and resolution>0:
                    if self.value is None or abs(value-self.value)>=resolution:
                        self.signalSync(value)
                else:
                    if value != self.value:
                        self.signalSync(value)
                self._timeoutManaged=self.timeout(180)
                self.updateFlags()

    def set(self, value):
        self.value=value

    def off(self):
        self.set(0)

    def min(self):
        if self._min is not None:
            self.set(self._min)

    def max(self):
        if self._max is not None:
            self.set(self._max)

    def signalSync(self, value=None):
        if value is not None:
            self._toReachValue=value
        self._pendingSync=True
        self.parent.signalSync()

    def isPendingSync(self, reset=False):
        if self._pendingSync and self._toReachValue is not None:
            if reset:
                self.clearSync()
            return True
        return False

    @property
    def toReachValue(self):
        return self._toReachValue

    def clearSync(self):
        self._pendingSync=False

    def clearSyncAndUpdateValue(self):
        self.clearSync()
        if self.toReachValue is not None:
            self.updateValue(self.toReachValue)
            self._stamp=time.time()


class MBIOValueDigital(MBIOValue):
    def __init__(self, parent: MBIOValues, name, default=None, writable=False, commissionable=False):
        super().__init__(parent, name=name, unit=15, default=default, writable=writable, commissionable=commissionable)

    def isDigital(self):
        return True

    def normalizeValue(self, value):
        if value is not None:
            if type(value) is bool:
                return value
            if type(value) is int:
                return bool(value)
            if value is None:
                return False
            return bool(value)
        return value

    def isOn(self):
        if self.value:
            return True
        return False

    def isOff(self):
        return not self.isOn()


class MBIOValueTrigger(MBIOValueDigital):
    def __init__(self, parent: MBIOValues, name, delay=15, commissionable=False):
        super().__init__(parent, name=name, default=False, commissionable=commissionable)
        self._triggerDelay=delay
        self._triggerNotified=False
        self._timeoutTrigger=0

    def hasManager(self):
        return True

    def trigger(self):
        if not self.isOn():
            self._triggerNotified=False
            self._timeoutTrigger=self.timeout(self._triggerDelay)
            self.updateValue(True)
            self.logger.warning('TRIGGER %s' % self)

    def manager(self):
        if self.isOn():
            if time.time()>=self._timeoutTrigger:
                self.updateValue(False)
                self.logger.warning("TRIGGERDONE %s" % self)

    def signalNotified(self):
        if self.isOn() and not self._triggerNotified:
            self._triggerNotified=True
            self.logger.debug('TRIGGER NOTIFIED %s' % self)


class MBIOValueDigitalWritable(MBIOValueDigital):
    def __init__(self, parent: MBIOValues, name, default=None, commissionable=False):
        super().__init__(parent, name=name, default=default, writable=True, commissionable=commissionable)
        self._toReachValue=None
        self._pendingSync=False

    def isManaged(self):
        """Return True if the value is writable and seems to be managed (value written not too long ago)"""
        try:
            if self.isValid() and time.time()<self._timeoutManaged:
                return True
            return False
        except:
            self.logger.exception('xx')

    @MBIOValueDigital.value.setter
    def value(self, value):
        if value is not None:
            if self.isEnabled():
                if self.isManual():
                    value=self._manual
                value=self.normalizeValue(value)
                if value!=self.value:
                    self.signalSync(value)
                self._timeoutManaged=self.timeout(180)
                self.updateFlags()

    def set(self, value):
        self.value=value

    def signalSync(self, value=None):
        if value is not None:
            self._toReachValue=value
        self._pendingSync=True
        self.parent.signalSync()

    def isPendingSync(self, reset=False):
        if self._pendingSync and self._toReachValue is not None:
            if reset:
                self.clearSync()
            return True
        return False

    @property
    def toReachValue(self):
        return self._toReachValue

    def clearSync(self):
        self._pendingSync=False

    def clearSyncAndUpdateValue(self):
        self.clearSync()
        if self.toReachValue is not None:
            self.updateValue(self.toReachValue)
            self._stamp=time.time()

    def toggle(self):
        self.value=not self.value

    def on(self):
        self.value=True

    def off(self):
        self.value=False


class MBIOValueMultistate(MBIOValue):
    def __init__(self, parent: MBIOValues, name, vmax, vmin=0, default=None, writable=False, commissionable=False):
        super().__init__(parent, name=name, unit=53, default=default, resolution=1.0, writable=writable, commissionable=commissionable)
        if vmin is None or vmin<0:
            vmin=0
        self.setRange(vmin, vmax)

    def isDigital(self):
        return False

    def isMultistate(self):
        return True

    def normalizeValue(self, value):
        value=super().normalizeValue(value)
        try:
            return int(value)
        except:
            pass
        return self._min


class MBIOValueMultistateWritable(MBIOValueMultistate):
    def __init__(self, parent: MBIOValues, name, vmax, vmin=0, default=None, commissionable=False):
        super().__init__(parent, name=name, vmax=vmax, vmin=vmin, default=default, writable=True, commissionable=commissionable)
        self._toReachValue=None
        self._pendingSync=False

    def isManaged(self):
        """Return True if the value is writable and seems to be managed (value written not too long ago)"""
        if self.isValid() and time.time()<self._timeoutManaged:
            return True
        return False

    @MBIOValueDigital.value.setter
    def value(self, value):
        if value is not None:
            if self.isEnabled():
                if self.isManual():
                    value=self._manual
                value=self.normalizeValue(value)
                if value!=self.value:
                    self.signalSync(value)
                self._timeoutManaged=self.timeout(180)
                self.updateFlags()

    def set(self, value):
        self.value=value

    def signalSync(self, value=None):
        if value is not None:
            self._toReachValue=value
        self._pendingSync=True
        self.parent.signalSync()

    def isPendingSync(self, reset=False):
        if self._pendingSync and self._toReachValue is not None:
            if reset:
                self.clearSync()
            return True
        return False

    @property
    def toReachValue(self):
        return self._toReachValue

    def clearSync(self):
        self._pendingSync=False

    def clearSyncAndUpdateValue(self):
        self.clearSync()
        if self.toReachValue is not None:
            self.updateValue(self.toReachValue)
            self._stamp=time.time()

    def next(self):
        if self.value is not None:
            self.value=self.value+1
        else:
            self.value=self._min

    def prev(self):
        if self.value is not None:
            self.value=self.value-1
        else:
            self.value=self._min


class MBIOValues(Items):
    def __init__(self, parent, prefix, logger):
        super().__init__(logger)
        self._parent=parent
        self._prefix=prefix
        self._items: list[MBIOValue]=[]
        self._itemByKey={}
        self._itemByName={}
        self._units=Units()
        self._hasWritableValue=False

    @property
    def parent(self):
        return self._parent

    @property
    def prefix(self):
        return self._prefix

    @property
    def units(self):
        return self._units

    def computeValueKeyFromName(self, name):
        if self._prefix:
            return '%s_%s' % (self._prefix, name)
        return name

    def getByKeyComputedFromName(self, name):
        return self.getByKey(self.computeValueKeyFromName(name))

    def getMBIO(self) -> MBIO:
        return self._parent.getMBIO()

    def item(self, key):
        item=super().item(key)
        if item is not None:
            return item

        item=self.getByKey(key)
        if item is not None:
            return item

        item=self.getByName(key)
        if item is not None:
            return item

        try:
            return self[key]
        except:
            pass

    def hasWritableValue(self):
        if self._hasWritableValue:
            return True
        return False

    def add(self, value: MBIOValue) -> MBIOValue:
        if isinstance(value, MBIOValue):
            super().add(value)
            self._itemByKey[value.key]=value
            self._itemByName[value.name]=value
            try:
                attr=getattr(self, value.name)
                if attr is not None:
                    self.logger.warning('possible value variable name [%s] conflict with values object' % value.name)
                else:
                    setattr(self, value.name, value)
            except:
                setattr(self, value.name, value)
            if value._writable:
                self._hasWritableValue=True
            try:
                self.getMBIO().registerValue(value)
            except:
                pass

    def getByKey(self, key):
        try:
            return self._itemByKey[key]
        except:
            pass

    def getByName(self, key):
        try:
            return self._itemByName[key]
        except:
            pass

    def __radd__(self, other):
        self.add(other)

    def signalSync(self):
        try:
            self._parent.signalSync()
        except:
            pass

    # Force return None if called with an unknown attribute (value name)
    def __getattr__(self, name):
        try:
            return getattr(self, name)
        except:
            pass
        return None


if __name__ == "__main__":
    pass



#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .mbio import MBIO
    from .gateway import MBIOGateway

import time
import threading
import math
from prettytable import PrettyTable

from .config import MBIOConfig
from .xmlconfig import XMLConfig

from .items import Items
from .value import MBIOValues, MBIOValue, MBIOConstant, MBIOValueWritable
from .value import MBIOValueDigital, MBIOValueDigitalWritable
from .value import MBIOValueMultistate, MBIOValueMultistateWritable

from pymodbus.constants import Endian
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.payload import BinaryPayloadBuilder


# BinaryPayloadEncoder is deprecated
# to be replaced with client.convert_from_registers() and client.convert_to_registers()
# https://github.com/pymodbus-dev/pymodbus/blob/dev/pymodbus/client/mixin.py

class MBIOModbusRegistersEncoder(object):
    def __init__(self, device, encoding=Endian.BIG):
        self._device=device
        self._encoder=BinaryPayloadBuilder(byteorder=encoding, wordorder=encoding)

    def set(self, vtype, value):
        try:
            vtype=vtype.lower()
            if vtype=='word':
                return self.word(value)
            elif vtype=='float32':
                return self.float32(value)
            elif vtype=='dword':
                return self.dword(value)
            elif vtype=='int':
                return self.int(value)
            elif vtype=='int16':
                return self.int(value)
            elif vtype=='int32':
                return self.int32(value)
            elif vtype=='skip':
                self.word(0)
                return
            elif vtype=='float16':
                return self.float16(value)
        except:
            pass

    def word(self, value):
        self._encoder.add_16bit_uint(value)

    def int(self, value):
        self._encoder.add_16bit_int(value)

    def int16(self, value):
        self._encoder.add_16bit_int(value)

    def float16(self, value):
        self._encoder.add_16bit_float(value)

    def float32(self, value):
        self._encoder.add_32bit_float(value)

    def int32(self, value):
        self._encoder.add_32bit_int(value)

    def dword(self, value):
        self._encoder.add_32bit_uint(value)

    def payload(self):
        try:
            # return self._encoder.build()
            return self._encoder.to_registers()
        except:
            pass

    def writeRegisters(self, start):
        return self._device.writeRegisters(start, self.payload())

    def writeRegistersIfChanged(self, start):
        return self._device.writeRegistersIfChanged(start, self.payload())


class MBIOModbusRegistersDecoder(object):
    def __init__(self, r, encoding=Endian.BIG):
        self._decoder=BinaryPayloadDecoder.fromRegisters(r, byteorder=encoding, wordorder=encoding)

    def get(self, vtype):
        try:
            vtype=vtype.lower()
            if vtype=='word':
                return self.word()
            elif vtype=='float32':
                return self.float32()
            elif vtype=='dword':
                return self.dword()
            elif vtype=='int':
                return self.int()
            elif vtype=='int16':
                return self.int()
            elif vtype=='int32':
                return self.int32()
            elif vtype=='skip':
                self.word()
                return
            elif vtype=='float16':
                return self.float16()
        except:
            pass

    def word(self):
        try:
            return self._decoder.decode_16bit_uint()
        except:
            pass

    def int(self):
        try:
            return self._decoder.decode_16bit_int()
        except:
            pass

    def int16(self):
        try:
            return self._decoder.decode_16bit_int()
        except:
            pass

    def float16(self):
        try:
            return self._decoder.decode_16bit_float()
        except:
            pass

    def float32(self):
        try:
            return self._decoder.decode_32bit_float()
        except:
            pass

    def int32(self):
        try:
            return self._decoder.decode_32bit_int()
        except:
            pass

    def dword(self):
        try:
            return self._decoder.decode_32bit_uint()
        except:
            pass


class MBIODevice(object):
    STATE_OFFLINE = 0
    STATE_PROBE = 1
    STATE_POWERON = 2
    STATE_PREOP = 3
    STATE_ONLINE = 4
    STATE_POWEROFF = 5
    STATE_ERROR = 6
    STATE_HALT = 7

    def buildKey(self, gateway, address):
        return '%s_mb%d' % (gateway.key, address)

    def __init__(self, gateway: MBIOGateway, address, xml: XMLConfig = None):
        # assert(isinstance(gateway, MBIOGateway))
        self._gateway=gateway
        self._address=int(address)
        self._key=self.buildKey(gateway, address)
        self._eventReset=threading.Event()
        self._eventHalt=threading.Event()
        self._state=self.STATE_OFFLINE
        self._timeoutState=0
        self._timeoutRefresh=0
        self._timeoutSync=0
        self._timeoutReSync=self.timeout(60)
        self._timeoutSlowManager=0

        # TODO: enable sync after bootup
        self._syncEnable=True
        self._pingRegisterIndex=None
        self._pingRegisterWithHoldingRegister=False
        self._vendor=None
        self._model=None
        self._version=None
        self._error=False
        self._conditionalRegisterWriteCount=0
        self._conditionalWrite=False
        self._config=MBIOConfig()

        self._countMsg=0
        self._countMsgErr=0
        self._stampState=0
        self._countProbe=0

        self._cache=None

        self._values=MBIOValues(self, self._key, self.logger)
        self._sysvalues=MBIOValues(self, '%s' % self.key, self.logger)

        self._sysComErr=MBIOValueDigital(self._sysvalues, 'comerr')
        # self._sysStateUptime=MBIOValue(self._sysvalues, 'stup')
        # self._sysMsgCount=MBIOValue(self._sysvalues, 'msg')
        # self._sysMsgErrCount=MBIOValue(self._sysvalues, 'msgerr')

        self._gateway.devices.add(self)
        self.onInit()
        self.load(xml)

    @property
    def dtype(self):
        return self.__class__

    def onInit(self):
        pass

    def onLoad(self, xml: XMLConfig):
        pass

    def normalizeResistor(self, r):
        # 0..1600 ohms as 0..10V equivalent
        # allow usage of the Digimat-3 conversion formulas
        if (r>=0 and r<=1600):
            return r/1600.0 * 10.0
        return None

    def pt100(self, r):
        f=self.normalizeResistor(r)
        if f is not None:
            f=0.0769632 * (f**3) + 0.2946681 * (f**2) + 6.98698436 * (f) - 3.59148152
            return f*550.0/10.0-50.0
        return 0

    def pt1000(self, r):
        f=self.normalizeResistor(r)
        if f is not None:
            f=0.00011683 * (f**3) + 0.01047814 * (f**2) + 1.90211508 * (f) - 9.82603617
            return f*200.0/10.0-50.0
        return 0

    def ni1000_tk5000(self, r):
        f=self.normalizeResistor(r)
        if f is not None:
            f=0.00354046 * (f**3) - 0.1557009 * (f**2) + 3.65814114 * (f) - 14.70477509
            return f*170.0/10.0-50.0
        return 0

    def ni1000(self, r):
        f=self.normalizeResistor(r)
        if f is not None:
            f=0.00229944 * (f**3) - 0.10339928 * (f**2) + 2.74480809 * (f) - 10.73874259
            return f*170.0/10.0-50.0
        return 0

    def r2t(self, stype, r):
        # self.logger.warning("r(%f) to %s" % (r, stype))
        if stype=='pt1000':
            return self.pt1000(r)
        if stype=='pt100':
            return self.pt100(r)
        if stype=='ni1000':
            return self.ni1000(r)
        if stype=='ni1000-5k':
            return self.ni1000_tk5000(r)
        if stype=='ohm':
            return r
        return r

    def calcDEWP(self, t, hr):
        if t is None or t<0 or t>60:
            return 0
        if hr is None or hr<1 or hr>100:
            return 0

        a=17.27
        b=237.7
        v1=a*t/(b+t)
        v1+=math.log10(hr/100.0)
        v2=a-v1
        if v2!=0:
            return b*v1/v2

        return 0

    def microsleep(self):
        # should release the GIL
        time.sleep(0)

        # Another way to do it
        # os.sched_yield()

    def load(self, xml: XMLConfig):
        if xml:
            try:
                if xml.isConfig('device'):
                    self.onLoad(xml)
            except:
                self.logger.exception('%s:%s:load()' % (self.__class__.__name__, self.key))

    @property
    def key(self):
        return self._key

    @property
    def address(self):
        return self._address

    @property
    def gateway(self) -> MBIOGateway:
        return self._gateway

    @property
    def parent(self):
        return self.gateway

    def getMBIO(self) -> MBIO:
        return self.gateway.getMBIO()

    @property
    def mbio(self):
        return self.getMBIO()

    @property
    def config(self) -> XMLConfig:
        return self._config

    @property
    def values(self):
        return self._values

    @property
    def sysvalues(self):
        return self._sysvalues

    def value(self, name, unit=0xff, default=None, writable=False, resolution=0.1, commissionable=False):
        key=self.values.computeValueKeyFromName(name)
        value=self.values.item(key)
        if value is None:
            if writable:
                value=MBIOValueWritable(self.values, name=name, unit=unit, default=default, resolution=resolution, commissionable=commissionable)
            else:
                value=MBIOValue(self.values, name=name, unit=unit, default=default, resolution=resolution, commissionable=commissionable)
        return value

    def constant(self, name, value, unit):
        key=self.values.computeValueKeyFromName(name)
        value=self.values.item(key)
        if value is None:
            value=MBIOConstant(self.values, name=name, value=value, unit=unit)
        return value

    def valueDigital(self, name, default=None, writable=False, commissionable=False):
        key=self.values.computeValueKeyFromName(name)
        value=self.values.item(key)
        if value is None:
            if writable:
                value=MBIOValueDigitalWritable(self.values, name=name, default=default, commissionable=commissionable)
            else:
                value=MBIOValueDigital(self.values, name=name, default=default, commissionable=commissionable)
        return value

    def valueMultistate(self, name, vmax, vmin=0, default=None, writable=False, commissionable=False):
        key=self.values.computeValueKeyFromName(name)
        value=self.values.item(key)
        if value is None:
            if writable:
                value=MBIOValueMultistateWritable(self.values, name=name, vmax=vmax, vmin=vmin, default=default, commissionable=commissionable)
            else:
                value=MBIOValueMultistate(self.values, name=name, vmax=vmax, vmin=vmin, default=default, commissionable=commissionable)
        return value

    @property
    def logger(self):
        return self._gateway.logger

    @property
    def client(self):
        return self._gateway.client

    @property
    def countMsg(self):
        return self._countMsg

    @property
    def countMsgErr(self):
        return self._countMsgErr

    def statetime(self):
        if self._stampState>0:
            return time.time()-self._stampState
        return 0

    def enableSync(self):
        self.logger.info('Enable SYNC on device %s' % self._key)
        self._syncEnable=True

    def setPingRegister(self, index, useInputRegister=True):
        self._pingRegisterIndex=index
        self._pingRegisterWithHoldingRegister=not useInputRegister

    def setPingInputRegister(self, index):
        return self.setPingRegister(index, True)

    def setPingHoldingRegister(self, index):
        return self.setPingRegister(index, False)

    def ping(self):
        if self._pingRegisterIndex is not None:
            self.logger.debug('Modbus pinging device %s' % self.key)
            self.gateway.checkIdleAfterSend()
            if self._pingRegisterWithHoldingRegister:
                if self.readHoldingRegisters(self._pingRegisterIndex) is not None:
                    self.gateway.signalMessageTransmission()
                    return True
            else:
                self.gateway.checkIdleAfterSend()
                if self.readInputRegisters(self._pingRegisterIndex) is not None:
                    self.gateway.signalMessageTransmission()
                    return True
        else:
            self.logger.debug('Generic pinging device %s' % self.key)
            self.gateway.checkIdleAfterSend()
            if self.gateway.ping(self.address):
                self.gateway.signalMessageTransmission()
                return True

        self.gateway.signalMessageTransmission()
        self.logger.error('Unable to ping device %s' % self.key)
        return False

    def probe(self):
        return self.gateway.probe(self.address)

    def timeout(self, delay):
        return time.time()+delay

    def isTimeout(self, t):
        if t is None or time.time()>=t:
            return True
        return False

    def isElapsed(self, t, delay):
        if time.time()>=t+delay:
            return True
        return False

    def isOnline(self):
        if self.state==self.STATE_ONLINE:
            return True
        return False

    def isHalted(self):
        if self.state==self.STATE_HALT:
            return True
        return False

    def isError(self):
        if self._error:
            return True
        return False

    def setError(self, state=True):
        if self._error!=state:
            self._error=state
            self._sysComErr.updateValue(state)
            for value in self.values:
                value.setError(state)
            if state:
                self.logger.warning('Calling device %s powerlost()' % self.key)
                self.powerlost()

    def reset(self):
        self._eventReset.set()
        self._countProbe=0

    def halt(self):
        self._eventHalt.set()

    def restartCommunication(self):
        try:
            self.logger.info('Restart device %s communication' % self.key)
            self.gateway.checkIdleAfterSend()
            r=self.client.diag_restart_communication(True, slave=self.address)
            self.gateway.signalMessageTransmission()
            if r and not r.isError():
                return True
        except:
            pass
        return False

    @property
    def vendor(self):
        return self._vendor

    @property
    def model(self):
        return self._model

    @property
    def version(self):
        return self._version

    def readDiscreteInputs(self, start, count=1):
        try:
            self.gateway.checkIdleAfterSend()
            self._countMsg+=1
            r=self.client.read_discrete_inputs(start, count, slave=self.address)
            self.gateway.signalMessageTransmission()
            if r and not r.isError():
                self.signalAlive()
                return r.bits
        except:
            pass

        self._countMsgErr+=1
        self.logger.error('<--readDiscretInputs %s:%d#%d' % (self.key, start, count))

    def readCoils(self, start, count=1):
        try:
            self.gateway.checkIdleAfterSend()
            self._countMsg+=1
            r=self.client.read_coils(start, count, slave=self.address)
            self.gateway.signalMessageTransmission()
            if r and not r.isError():
                self.signalAlive()
                return r.bits
        except:
            pass

        self._countMsgErr+=1
        self.logger.error('<--readCoils %s:%d#%d' % (self.key, start, count))

    def writeCoils(self, start, data):
        if data is not None:
            data=self.ensureArray(data)
            try:
                self.gateway.checkIdleAfterSend()
                self._countMsg+=1
                r=self.client.write_coils(start, data, slave=self.address)
                self.gateway.signalMessageTransmission()
                if r and not r.isError():
                    self.signalAlive()
                    return True
            except:
                pass

        self._countMsgErr+=1
        self.logger.error('<--writeCoils %s:%d' % (self.key, start))
        return False

    def readInputRegisters(self, start, count=1):
        try:
            self.gateway.checkIdleAfterSend()
            self._countMsg+=1
            r=self.client.read_input_registers(start, count, slave=self.address)
            self.gateway.signalMessageTransmission()
            if r and not r.isError():
                self.signalAlive()
                return r.registers
        except:
            pass

        self._countMsgErr+=1
        self.logger.error('<--readInputRegisters %s:%d(%X)#%d' % (self.key, start, start, count))

    def cacheDisable(self):
        self._cache=None

    def cacheReset(self):
        self._cache={'stamp': time.time()}

    def cacheEnable(self):
        if self._cache is None:
            self.cacheReset()

    def cacheAge(self):
        try:
            return time.time()-self._cache['stamp']
        except:
            pass
        return 0

    def cachePrune(self, age):
        if self.cacheAge()>age:
            self.cacheReset()

    def cache(self, dtype):
        if self._cache is None:
            return None

        if dtype:
            try:
                return self._cache[dtype]
            except:
                pass
            self._cache[dtype]={}
            return self._cache[dtype]

    def cacheGet(self, dtype, start, count):
        cache=self.cache(dtype)
        try:
            r=[]
            for n in range(start, start+count):
                r.append(cache[n])
            return r
        except:
            pass

    def cacheSet(self, dtype, start, r):
        cache=self.cache(dtype)
        try:
            n=start
            for data in r:
                cache[n]=data
                n+=1
        except:
            pass

    def cacheReadInputRegisters(self, start, count=1, block=32, base=0):
        self.cacheEnable()

        dtype='ir'
        data=self.cacheGet(dtype, start, count)
        if data:
            return data

        if start>=base:
            n0=((start-base) // block) * block
            nread=0
            while nread<count:
                # self.logger.debug('<--cacheReadInputRegisters(%d, %d)' % (n0+base, block))
                r=self.readInputRegisters(base+n0, block)
                if not r:
                    break
                self.cacheSet(dtype, n0+base, r)
                nread+=block
                n0+=block

            return self.cacheGet(dtype, start, count)

    def readHoldingRegisters(self, start, count=1):
        try:
            self.gateway.checkIdleAfterSend()
            self._countMsg+=1
            r=self.client.read_holding_registers(start, count, slave=self.address)
            self.gateway.signalMessageTransmission()
            if r and not r.isError():
                self.signalAlive()
                return r.registers
        except:
            pass
        self.logger.error('<--readHoldingRegisters %s:%d(%X)#%d' % (self.key, start, start, count))

    def cacheReadHoldingRegisters(self, start, count=1, block=32, base=0):
        self.cacheEnable()

        dtype='hr'
        data=self.cacheGet(dtype, start, count)
        if data:
            return data

        if start>=base:
            n0=((start-base) // block) * block
            nread=0
            while nread<count:
                # self.logger.debug('<--cacheReadHoldingRegisters(%d, %d)' % (n0+base, block))
                r=self.readHoldingRegisters(base+n0, block)
                if not r:
                    break
                self.cacheSet(dtype, n0+base, r)
                nread+=block
                n0+=block

            return self.cacheGet(dtype, start, count)

    def ensureArray(self, data):
        if data is not None:
            try:
                data[0]
            except:
                # convert to array
                data=[data]
        return data

    def writeRegisters(self, start, data):
        if data is not None:
            data=self.ensureArray(data)
            try:
                self.gateway.checkIdleAfterSend()
                self._countMsg+=1
                r=self.client.write_registers(start, data, slave=self.address)
                self.gateway.signalMessageTransmission()
                if r and not r.isError():
                    self.signalAlive()
                    return True
            except:
                pass

        self._countMsgErr+=1
        self.logger.error('<--writeRegisters %s:%d(0x%X) %s' % (self.key, start, start, str(data)))
        return False

    def writeHoldingRegisters(self, start, data):
        return self.writeRegisters(start, data)

    def resetConditionalRegisterWriteCount(self):
        self._conditionalRegisterWriteCount=0
        self._conditionalWrite=False

    def getConditionalRegisterWriteCount(self):
        return self._conditionalRegisterWriteCount

    def checkConditionalWriteMark(self, reset=True):
        if self.getConditionalRegisterWriteCount()>0:
            if reset:
                self._conditionalWrite=False
            return True
        return False

    def resetConditionalWriteMark(self):
        self.checkConditionalWriteMark(True)

    def markConditionalWrite(self):
        self._conditionalRegisterWriteCount+=1
        self._conditionalWrite=True

    def writeRegistersIfChanged(self, start, data):
        if data is not None:
            data=self.ensureArray(data)
            size=len(data)
            r=self.readHoldingRegisters(start, size)
            if r:
                for n in range(size):
                    if r[n]!=data[n]:
                        # mismatch, we have to write it
                        self.logger.warning('<--writeRegisters[CauseChanged] %s:%d(0x%X) from %s to %s' %
                            (self.key, start, start, str(r), str(data)))
                        if self.writeRegisters(start, data):
                            self.markConditionalWrite()
                            return True
                        return False
                # Nothing to change, so it's a success
                return True
        return False

    def writeRegisterBitIfChanged(self, start, bit, state):
        r=self.readHoldingRegisters(start, 1)
        if r:
            v0=r[0]
            if state:
                v1 = v0 | (0x1 << bit)
            else:
                v1 = (v0 & (~(0x1 << bit))) & 0xffff

            if v1!=v0:
                # mismatch, we have to write it
                self.logger.warning('<--writeRegisterBitIfChanged[CauseChanged] %s:%d(0x%X) from %s to %s' %
                    (self.key, start, start, bin(v0), bin(v1)))
                if self.writeRegisters(start, v1):
                    self.markConditionalWrite()
                    return True
                return False
            # Nothing to change, so it's a success
            return True
        return False

    def setRegisterBitIfChanged(self, start, bit):
        return self.writeRegisterBitIfChanged(start, bit, 1)

    def clearRegisterBitIfChanged(self, start, bit):
        return self.writeRegisterBitIfChanged(start, bit, 0)

    def decoderFromRegisters(self, r, encoding=Endian.BIG) -> MBIOModbusRegistersDecoder:
        if r:
            try:
                return MBIOModbusRegistersDecoder(r, encoding)
            except:
                pass

    def encoder(self, encoding=Endian.BIG) -> MBIOModbusRegistersEncoder:
        return MBIOModbusRegistersEncoder(self, encoding)

    def poweron(self):
        # to be overriden
        return True

    def poweronsave(self):
        # to be overriden
        # called if any conditional registers write were done during poweron()
        self.logger.debug('Device %s poweron phase has changed some config data!' % (self.key))

    def isreadytorun(self):
        # to be overriden
        return True

    def poweroff(self):
        # to be overriden
        return True

    def powerlost(self):
        # to be overriden
        # TODO: not a good name
        return True

    def sync(self):
        # self.logger.warning('Fallback %s:sync()' % self.key)
        for value in self.values:
            try:
                if value.isPendingSync():
                    self.logger.debug('Fallback default SYNC: clearSync(%s)' % (value))
                    value.clearSync()
            except:
                pass
        return True

    def refresh(self):
        # to be overriden
        return True

    def resync(self):
        if self.values.hasWritableValue():
            self.logger.debug('RESYNC device %s' % self.key)
            for value in self.values:
                # Force rewrite only values that are managed by the CPU
                if value.isWritable() and value.isManaged():
                    self.logger.warning('RESYNC FORCE REWRITE %s' % value)
                    value.signalSync()
            self._timeoutReSync=self.timeout(180)
            self.signalSync()

    def run(self):
        if self.isPendingRefresh(True):
            # self.logger.debug('REFRESH %s#%d' % (self.model, self.address))
            timeout=self.refresh()
            # self.logger.debug('REFRESHDONE %s#%d' % (self.model, self.address))
            if timeout is None:
                timeout=5.0
            self._timeoutRefresh=self.timeout(timeout)

        if self._syncEnable and self.isPendingSync(True):
            self.logger.debug('SYNC device %s' % self.key)
            self.sync()
            # self.logger.debug('SYNCDONE device %s' % self.key)
            self.signalRefresh()
            self._timeoutReSync=self.timeout(180)
        else:
            if self.isTimeout(self._timeoutReSync):
                self.resync()

        return True

    def signalAlive(self):
        if self._state==self.STATE_ONLINE:
            # self.logger.debug("SignalAlive(%s)!" % self.key)
            self._timeoutState=self.timeout(15)

    @property
    def state(self):
        return self._state

    def statestr(self):
        states=['OFFLINE', 'PROBING', 'POWERON', 'PREOP', 'ONLINE', 'POWEROFF', 'ERROR', 'HALT']
        try:
            return states[self.state]
        except:
            pass
        return 'UNKNOWN:%d' % self._state

    def richstatestr(self):
        states=['[red bold]OFFLINE[/red bold]',
                '[bold blue]PROBING[/bold blue]',
                '[bold blue]POWERON[/bold blue]',
                '[bold blue]PREOP[/bold blue]',
                '[bold green]ONLINE[/bold green]',
                '[bold blue]POWEROFF[/bold blue]/',
                '[bold red]ERROR[/bold red]',
                '[bold red]HALT[/bold red]']
        try:
            return states[self.state]
        except:
            pass
        return 'UNKNOWN:%d' % self._state

    def updateValuesFlags(self):
        for value in self.values:
            value.updateFlags()

    def setState(self, state, timeout=0):
        if state!=self._state:
            self._state=state
            self._stampState=time.time()
            self.logger.debug('Changing device %s state to %d:%s (T=%ds)' % (self.key,  state, self.statestr(), timeout))
            self._timeoutState=self.timeout(timeout)
            self.updateValuesFlags()

    def slowManager(self):
        # now=time.time()
        # self._sysStateUptime.updateValue(now-self._stampState)
        # self._sysMsgCount.updateValue(self._countMsg)
        # self._sysMsgErrCount.updateValue(self._countMsgErr)
        pass

    def manager(self):
        if time.time()>=self._timeoutState:
            self.logger.debug("%s state timeout!" % self.key)
            timeout=True
        else:
            timeout=False

        if self.isTimeout(self._timeoutSlowManager):
            self.slowManager()
            self._timeoutSlowManager=self.timeout(5)

        self._sysComErr.updateValue(self.isError())

        # ----------------------------------------------------
        if self._state==self.STATE_ONLINE:
            if timeout:
                if not self.ping():
                    self.setState(self.STATE_ERROR, 5)
                    return
            if self._eventReset.is_set():
                self.setState(self.STATE_OFFLINE, 1)
                return

            if self._eventHalt.is_set():
                self.setState(self.STATE_POWEROFF)
                return

            self.run()
            self.setError(False)
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_OFFLINE:
            if timeout:
                self._eventReset.clear()
                # FIXME: not always supported
                self.restartCommunication()
                self.setState(self.STATE_PROBE)
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_PROBE:
            self._countProbe+=1
            data=self.probe()
            if data:
                try:
                    if data['vendor']:
                        self._vendor=data['vendor']
                except:
                    pass
                try:
                    if data['model']:
                        self._model=data['model']
                except:
                    pass
                try:
                    if data['version']:
                        self._version=data['version']
                except:
                    pass

            if self.ping():
                self.setState(self.STATE_POWERON)
                return

            if self._countProbe>3:
                self.setState(self.STATE_ERROR, 60)
            else:
                self.setState(self.STATE_ERROR, 5)
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_POWERON:
            self._countProbe=0
            self.resetConditionalRegisterWriteCount()
            self.logger.debug('Calling device %s poweron()' % self.key)
            if self.poweron():
                if self.getConditionalRegisterWriteCount()>0:
                    self.microsleep()
                    self.logger.debug('Saving updated poweron config of device %s' % self.key)
                    self.poweronsave()
                self.signalRefresh()
                self.setState(self.STATE_PREOP, 15)
                self.resync()
                return

            self.setState(self.STATE_ERROR, 15)
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_PREOP:
            if timeout:
                self.setState(self.STATE_ERROR, 15)
                return
            if self.isreadytorun():
                self.logger.debug('Device %s is now ready to run' % self.key)
                self.setState(self.STATE_ONLINE, 5)
                self._countMsg=0
                self._countMsgErr=0
                return
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_POWEROFF:
            self.logger.debug('Calling device %s poweroff()' % self.key)
            self.poweroff()
            if self._eventHalt.is_set():
                self.setState(self.STATE_HALT)
                return

            self.setState(self.STATE_OFFLINE, 1)
            return

        # ----------------------------------------------------
        elif self._state==self.STATE_ERROR:
            self.setError(True)
            if self._eventHalt.is_set():
                self.setState(self.STATE_HALT)
                return
            if self._eventReset.is_set() or timeout:
                self.setState(self.STATE_OFFLINE, 1)
                return

        # ----------------------------------------------------
        elif self._state==self.STATE_HALT:
            self._eventHalt.clear()
            self.setError(True)
            if self._eventReset.is_set():
                self.setState(self.STATE_OFFLINE, 1)
                return

        # ----------------------------------------------------
        else:
            # self.logger.error('unkown state %d' % self._state)
            self.setState(self.STATE_ERROR, 5)

    def __repr__(self):
        return '%s(%s=%s/%s, %s#%d)' % (self.__class__.__name__, self.key, self.vendor, self.model,
                self.statestr(), self.statetime())

    def richstr(self):
        return '[yellow]%s[/yellow]([bold]%s[/bold]=%s/%s, %s#%ds, %d/%d msg)' % (self.__class__.__name__,
                    self.key, self.vendor, self.model, self.richstatestr(),
                    self.statetime(), self.countMsg, self.countMsgErr)

    def dump(self):
        t=PrettyTable()
        t.field_names=['Property', 'Value']
        t.align='l'

        t.add_row(['key', self.key])
        t.add_row(['state', self.statestr()])

        for value in self.values:
            t.add_row([value.key, str(value)])
        for value in self._sysvalues:
            t.add_row([value.key, str(value)])

        print(t.get_string())

    def registerValue(self, value):
        self.gateway.parent.registerValue(value)

    def signalSync(self, delay=0):
        timeout=self.timeout(delay)
        if self._timeoutSync is None or timeout<self._timeoutSync:
            self._timeoutSync=timeout

    def isPendingSync(self, reset=True):
        if self._timeoutSync is not None:
            if self.isTimeout(self._timeoutSync):
                if reset:
                    self._timeoutSync=None
                return True
        return False

    def signalRefresh(self, delay=0.0):
        timeout=self.timeout(delay)
        if self._timeoutRefresh is None or timeout<self._timeoutRefresh:
            self._timeoutRefresh=timeout

    def isPendingRefresh(self, reset=True):
        if self._timeoutRefresh is not None:
            if self.isTimeout(self._timeoutRefresh):
                if reset:
                    self._timeoutRefresh=None
                return True
        return False

    def off(self):
        pass

    def auto(self):
        for value in self.values:
            value.auto()

    # def manual(self):
        # for value in self.values:
            # value.manual()

    def __getitem__(self, key):
        return self.values[key]


class MBIODeviceGeneric(MBIODevice):
    def onInit(self):
        self._vendor='Generic'
        self._model='Base'
        self.config.set('refreshperiod', 10)
        self.config.set('refresh', None)

    def onLoad(self, xml: XMLConfig):
        self._pingRegisterIndex=xml.getInt('pingInputRegister')
        if self._pingRegisterIndex is None:
            self._pingRegisterIndex=xml.getInt('pingHoldingRegister')
            self._pingRegisterWithHoldingRegister=True

        items=xml.child('values')
        if items:
            for item in items.children('value'):
                writable=item.getBool('writable')
                if item.getBool('digital'):
                    self.valueDigital(item.get('name'), writable=writable)
                else:
                    self.value(item.get('name'), writable=writable)

            for item in items.children('valuedigital'):
                writable=item.getBool('writable')
                self.valueDigital(item.get('name'), writable=writable)

        items=xml.child('refresh')
        if items:
            self.config.set('refreshperiod', items.getInt('period'))
            self.config.set('refresh', items)

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def refresh(self):
        self.logger.warning('%s:refresh()' % self.__class__.__name__)
        items=self.config.refresh
        if items:
            for item in items.children():
                start=item.getInt('start', 0)
                count=item.getInt('count', 1)
                r=None
                if item.tag=='holdingregisters':
                    r=self.readHoldingRegisters(start, count)
                elif item.tag=='inputregisters':
                    r=self.readInputRegisters(start, count)
                decoder=self.decoderFromRegisters(r)
                if decoder:
                    for data in item.children():
                        value=self.values.getByKeyComputedFromName(data.get('target'))
                        if value:
                            try:
                                vdata=decoder.get(data.tag)
                                f=data.getFloat('multiplyby')
                                if f:
                                    vdata*=f
                                f=data.getFloat('divideby')
                                if f:
                                    vdata/=f
                                f=data.getFloat('offset')
                                if f:
                                    vdata+=f
                                value.updateValue(vdata)
                            except:
                                pass
        return self.config.refreshperiod

    def sync(self):
        pass


class MBIODevices(Items):
    def __init__(self, logger):
        super().__init__(logger)
        self._items: list[MBIODevices]=[]
        self._itemByKey={}
        self._itemByAddress={}

    def item(self, key):
        item=self.getByAddress(key)
        if item:
            return item

        item=self.getByKey(key)
        if item:
            return item

    def add(self, item: MBIODevice) -> MBIODevice:
        if isinstance(item, MBIODevice):
            super().add(item)
            self._itemByKey[item.key]=item
            self._itemByAddress[item.address]=item

    def getByKey(self, key):
        try:
            return self._itemByKey[key]
        except:
            pass

    def getByAddress(self, address):
        try:
            return self._itemByAddress[address]
        except:
            pass

    def stop(self):
        for item in self._items:
            item.stop()

    def reset(self):
        for item in self._items:
            item.reset()

    def halt(self):
        for item in self._items:
            item.halt()

    def resetHalted(self):
        for item in self._items:
            if item.isHalted():
                item.reset()

    # def dump(self):
        # if not self.isEmpty():
            # t=PrettyTable()
            # t.field_names=['#', 'Address', 'Key', 'Host', 'Open']
            # t.align='l'
            # for item in self._items:
                # t.add_row([self.index(item), item.key, item.host, item.isOpen()])

        # print(t.get_string(sortby="Key"))


if __name__ == "__main__":
    pass



#!/bin/python

from __future__ import annotations

from .device import MBIODevice
# from prettytable import PrettyTable



# from pymodbus.constants import Endian

from .xmlconfig import XMLConfig


class MBIODeviceSensortecELD(MBIODevice):
    def onInit(self):
        self._vendor='Sensortec'
        self._model='EL-D'
        self.setPingInputRegister(39)
        self.t=self.value('t', unit='C', commissionable=True)
        self.hr=self.value('hr', unit='%', resolution=1)
        self.dewp=self.value('dewp', unit='C')

        self.config.set('icons', False)
        self.config.set('occupancy', False)
        self.config.set('fan', False)
        self.config.set('setpoint', False)
        self.config.set('temperature', False)
        self.config.set('toffset', 0)
        self.config.set('hygrometry', False)
        self.config.set('iaq', False)
        self.config.set('co2', False)
        self.config.set('cov', False)
        self.config.set('iaq', False)
        self.config.set('dark', False)

    def probe(self):
        self.logger.debug('Probing device address %d' % self.address)
        data={}
        r=self.readInputRegisters(39, 1)
        if r:
            v={0x20: 'EL-D-MB', 0x21: 'EL-D-TH-MB', 0x22: 'EL-D-THC-MB', 0x23: 'EL-D-THCV-MB'}
            data['model']=v.get(r[0], '?')

        r=self.readInputRegisters(15, 4)
        if r:
            data['version']='%d.%d' % ((r[0] >> 8), (r[0] & 0xff))
            data['serial']='%d' % (r[1]+(r[2] << 16))
            data['type']='%d' % (r[3])

        return data

    def onLoad(self, xml: XMLConfig):
        self.config.dark=xml.match('theme', 'dark')
        self.config.brightness=xml.getInt('brightness', 100, vmin=0, vmax=100)
        self.config.backlight=xml.getInt('backlight', 15, vmin=5, vmax=15)
        self.config.standby=xml.getInt('standby', 60, vmin=1, vmax=60)
        self.config.proximity=xml.getInt('proximity', 100, vmin=0, vmax=100)

        item=xml.child('hygrometry')
        if item:
            self.config.hygrometry=True
            self.config.set('hrhidden', item.getBool('hidden', False))

        item=xml.child('temperature')
        if item:
            self.config.temperature=True
            self.config.update('toffset',  item.getFloat('offset', 0, vmin=-3.0, vmax=3.0))
            self.config.set('thidden', item.getBool('hidden', False))

        item=xml.child('co2')
        if item:
            self.config.co2=True
            self.config.iaq=True
            self.config.set('co2warning',  item.getInt('warning', 800))
            self.config.set('co2alarm', item.getInt('alarm', 1200))
            self.config.set('co2hidden', item.getBool('hidden', False))
            self.co2=self.value('co2', unit='ppm', resolution=10)

        item=xml.child('cov')
        if item:
            self.config.cov=True
            self.config.iaq=True
            self.config.set('covwarning',  item.getInt('warning', 100))
            self.config.set('covalarm', item.getInt('alarm', 300))
            self.config.set('covhidden', item.getBool('hidden', False))
            self.cov=self.value('cov', unit='ppb', resolution=10)
            self.ecov=self.value('ecov', unit='ppm', resolution=10)

        if self.config.iaq:
            self.iaqwarning=self.valueDigital('iaqwarning')
            self.iaqalarm=self.valueDigital('iaqalarm')
            self.iaq=self.value('iaq', unit='', resolution=1)

        item=xml.child('Setpoint')
        if item:
            self.config.setpoint=True
            self.config.set('spmodeabsolute', not item.getBool('relative', False))
            self.config.set('spzero', item.getFloat('zero', 22.0))
            self.config.set('sprange', item.getFloat('range', 3.0))
            self.config.set('spstep', item.getFloat('step', 0.5, vmin=0.1, vmax=1.0))
            self.config.set('sphidden', item.getBool('hidden', False))
            self.spT=self.value('spt', unit='C', writable=True, commissionable=True)
            self.spTX=self.value('sptx', unit='C')

        item=xml.child('Fan')
        if item:
            self.config.fan=True
            self.config.set('fanstages', item.getInt('stages', 3, vmin=0, vmax=3))
            self.fanAuto=self.valueDigital('fanauto', writable=True)
            self.fanSpeed=self.value('fanspeed', unit='', writable=True, commissionable=True)

        if xml.child('Icons'):
            self.config.icons=True
            self.cooling=self.valueDigital('cooling', writable=True, default=False)
            self.heating=self.valueDigital('heating', writable=True, default=False)

        if xml.child('Occupancy'):
            self.config.occupancy=True
            self.occ=self.valueDigital('occ', writable=True, default=False)

    def poweron(self):
        # Communication
        self.writeRegistersIfChanged(11, [19200, 0])

        if self.config.dark:
            self.writeRegistersIfChanged(21, 2)
        else:
            self.writeRegistersIfChanged(21, 1)

        self.writeRegistersIfChanged(22, self.config.brightness)
        self.writeRegistersIfChanged(23, self.config.backlight)
        self.writeRegistersIfChanged(24, self.config.standby)
        self.writeRegistersIfChanged(38, int(self.config.proximity/10.0))

        v=0
        if self.config.setpoint and not self.config.sphidden:
            v |= (0x1)
        if self.config.fan:
            v |= (0x1 << 1)
        if self.config.occupancy:
            v |= (0x1 << 2)
        if self.config.temperature and not self.config.thidden:
            v |= (0x1 << 3)
        if self.config.hygrometry and not self.config.hrhidden:
            v |= (0x1 << 4)
        if self.config.icons:
            v |= (0x1 << 5)
        if self.config.iaq and (not self.config.co2hidden or not self.config.covhidden):
            v |= (0x1 << 6)
        self.writeRegistersIfChanged(20, v)

        self.writeRegistersIfChanged(14, self.config.toffset*100)

        if (self.config.co2 and not self.config.co2hidden) and (self.config.cov and not self.config.covhidden):
            self.writeRegistersIfChanged(32, 3)
        elif self.config.co2 and not self.config.co2hidden:
            self.writeRegistersIfChanged(32, 2)
        elif self.config.cov and not self.config.covhidden:
            self.writeRegistersIfChanged(32, 1)
        else:
            self.writeRegistersIfChanged(32, 0)

        if self.config.co2:
            self.writeRegistersIfChanged(35, self.config.co2warning)
            self.writeRegistersIfChanged(36, self.config.co2alarm)

        if self.config.cov:
            self.writeRegistersIfChanged(33, self.config.covwarning)
            self.writeRegistersIfChanged(34, self.config.covalarm)

        if self.config.iaq:
            # TODO: should be writeCoilsIfChanged
            self.writeCoils(507, 0)

        # iaq histeresis (ppm/ppb)
        self.writeRegistersIfChanged(37, 100)

        if self.config.setpoint:
            self.writeRegistersIfChanged(26, int(self.config.spstep*100.0))
            if self.config.spmodeabsolute:
                self.writeRegistersIfChanged(25, 0)
                self.writeRegistersIfChanged(27, int((self.config.spzero-self.config.sprange)*100.0))
                self.writeRegistersIfChanged(28, int((self.config.spzero+self.config.sprange)*100.0))
            else:
                self.writeRegistersIfChanged(25, 1)
                encoder=self.encoder()
                v=-int(self.config.sprange*100.0)
                encoder.int16(v)
                encoder.writeRegistersIfChanged(29)
                encoder=self.encoder()
                v=int(self.config.sprange*100.0)
                encoder.int16(v)
                encoder.writeRegistersIfChanged(30)

        # Fan
        if self.config.fan:
            if self.config.fanstages==0:
                self.writeRegistersIfChanged(31, 0x0102)
            elif self.config.fanstages==1:
                self.writeRegistersIfChanged(31, 0x0104)
            elif self.config.fanstages==2:
                self.writeRegistersIfChanged(31, 0x0108)
            else:
                self.writeRegistersIfChanged(31, 0x0110)
        else:
            self.writeRegistersIfChanged(31, 0x0110)

        return True

    def poweroff(self):
        return True

    def refresh(self):
        r=self.readInputRegisters(0, 8)
        if r:
            decoder=self.decoderFromRegisters(r)
            self.t.updateValue(decoder.word()/100.0)
            self.hr.updateValue(decoder.word()/100.0)
            self.dewp.updateValue(self.calcDEWP(self.t.value, self.hr.value))
            vsp=decoder.int16()/100.0
            if self.config.setpoint:
                self.spT.updateValue(vsp)
                if self.config.spmodeabsolute:
                    v=self.spT.value-self.config.spzero
                    self.spTX.updateValue(v)
                else:
                    v=self.config.spzero+self.spT.value
                    self.spTX.updateValue(v)

            vco2=decoder.word()
            status=decoder.word()
            vcov=decoder.word()
            vecov=decoder.word()
            viaq=decoder.word()
            if self.config.co2:
                self.co2.updateValue(vco2)
            if self.config.cov:
                self.cov.updateValue(vcov)
                self.ecov.updateValue(vecov)
            if self.config.iaq:
                self.iaq.updateValue(viaq)
                self.iaqwarning.updateValue(status & (0x1 << 5))
                self.iaqalarm.updateValue(status & (0x1 << 6))
            if self.config.occupancy:
                self.occ.updateValue(status & (0x1 << 1))
            if self.config.fan:
                self.fanAuto.updateValue(status & (0x1 << 8))
                speed=0
                if status & (0x1 << 10):
                    speed=1
                elif status & (0x1 << 11):
                    speed=2
                elif status & (0x1 << 12):
                    speed=3
                self.fanSpeed.updateValue(speed)
            if self.config.icons:
                self.heating.updateValue(status & (0x1 << 14))
                self.cooling.updateValue(status & (0x1 << 15))

        return 5.0

    def sync(self):
        if self.config.setpoint:
            value=self.spT
            if value.isPendingSync():
                if self.config.spmodeabsolute:
                    if self.writeRegisters(2, int(value.toReachValue*100)):
                        value.clearSync()
                else:
                    encoder=self.encoder()
                    encoder.int16(int(value.toReachValue*100))
                    if encoder.writeRegisters(2):
                        value.clearSync()

        if self.config.occupancy:
            value=self.occ
            if value.isPendingSync():
                if self.writeCoils(501, value.toReachValue):
                    value.clearSync()

        if self.config.fan:
            value=self.fanAuto
            if value.isPendingSync():
                if self.writeCoils(508, value.toReachValue):
                    value.clearSync()

            value=self.fanSpeed
            if value.isPendingSync():
                if value.toReachValue==0:
                    if self.writeCoils(509, 1):
                        value.clearSync()
                elif value.toReachValue==1:
                    if self.writeCoils(510, 1):
                        value.clearSync()
                elif value.toReachValue==2:
                    if self.writeCoils(511, 1):
                        value.clearSync()
                elif value.toReachValue==3:
                    if self.writeCoils(512, 1):
                        value.clearSync()

        if self.config.icons:
            value=self.heating
            if value.isPendingSync():
                if self.writeCoils(514, value.toReachValue):
                    value.clearSyncAndUpdateValue()
            value=self.cooling
            if value.isPendingSync():
                if self.writeCoils(515, value.toReachValue):
                    value.clearSyncAndUpdateValue()

    def reboot(self):
        if self.writeRegisters(99, 1):
            return True
        return False


class MBIODeviceSensortecEL(MBIODevice):
    def onInit(self):
        self._vendor='Sensortec'
        self._model='EL'
        self.setPingInputRegister(20)
        self.t=self.value('t', unit='C', commissionable=True)

        self.config.set('temperature', False)
        self.config.set('toffset', 0)
        self.config.set('hygrometry', False)
        self.config.set('iaq', False)
        self.config.set('co2', False)
        self.config.set('cov', False)

    def probe(self):
        self.logger.debug('Probing device address %d' % self.address)
        data={}
        r=self.readInputRegisters(16, 6)
        if r:
            data['version']='%d.%d' % ((r[0] >> 8), (r[0] & 0xff))
            data['serial']='%d' % (r[1]+(r[2] << 16))
            data['type']='%d' % (r[4])
            model='EL-'
            caps=r[5]
            if caps & (0x1<<0):
                model+='T'
            if caps & (0x1<<1):
                model+='H'
            if caps & (0x1<<2):
                model+='C'
            if caps & (0x1<<3):
                model+='V'
            data['model']=model

        return data

    def onLoad(self, xml: XMLConfig):
        item=xml.child('temperature')
        if item:
            self.config.temperature=True
            self.config.update('toffset',  item.getFloat('offset', 0, vmin=-2.0, vmax=2.0))

        item=xml.child('hygrometry')
        if item:
            self.config.hygrometry=True
            self.hr=self.value('hr', unit='%', resolution=1)
            self.dewp=self.value('dewp', unit='C')

        item=xml.child('co2')
        if item:
            self.config.co2=True
            self.config.iaq=True
            self.config.set('co2warning',  item.getInt('warning', 800))
            self.config.set('co2alarm', item.getInt('alarm', 1200))
            self.co2=self.value('co2', unit='ppm', resolution=10)

        item=xml.child('cov')
        if item:
            self.config.cov=True
            self.config.iaq=True
            self.config.set('covwarning',  item.getInt('warning', 100))
            self.config.set('covalarm', item.getInt('alarm', 300))
            self.cov=self.value('cov', unit='ppb', resolution=10)
            self.ecov=self.value('ecov', unit='ppm', resolution=10)

        if self.config.iaq:
            self.iaqwarning=self.valueDigital('iaqwarning')
            self.iaqalarm=self.valueDigital('iaqalarm')
            self.iaq=self.value('iaq', unit='', resolution=1)

    def poweron(self):
        # Communication
        self.writeRegistersIfChanged(11, [19200, 0])

        self.writeRegistersIfChanged(13, self.config.toffset*100)

        if self.config.co2:
            self.writeRegistersIfChanged(24, self.config.co2warning)
            self.writeRegistersIfChanged(25, self.config.co2alarm)

        if self.config.cov:
            self.writeRegistersIfChanged(22, self.config.covwarning)
            self.writeRegistersIfChanged(23, self.config.covalarm)

        if self.config.iaq:
            # iaq histeresis ppm/ppq
            self.writeRegistersIfChanged(26, 50)

        return True

    def poweroff(self):
        return True

    def refresh(self):
        r=self.readInputRegisters(0, 8)
        if r:
            decoder=self.decoderFromRegisters(r)
            self.t.updateValue(decoder.word()/100.0+self.config.toffset)
            vhr=decoder.word()/100.0
            if self.config.hygrometry:
                self.hr.updateValue(vhr)
                self.dewp.updateValue(self.calcDEWP(self.t.value, self.hr.value))

            vsp=decoder.int16()/100.0
            vco2=decoder.word()
            status=decoder.word()
            vcov=decoder.word()
            vecov=decoder.word()
            viaq=decoder.word()
            if self.config.co2:
                self.co2.updateValue(vco2)
            if self.config.cov:
                self.cov.updateValue(vcov)
                self.ecov.updateValue(vecov)
            if self.config.iaq:
                self.iaq.updateValue(viaq)
                self.iaqwarning.updateValue(status & (0x1 << 9))
                self.iaqalarm.updateValue(status & (0x1 << 10))

        return 5.0

    def sync(self):
        pass

    def reboot(self):
        if self.writeRegisters(19, 1):
            return True
        return False
    pass


if __name__ == "__main__":
    pass
